# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import glob
import os
import sys
from typing import List, NamedTuple


COPYRIGHT_LINES = [
    "Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.",
    "",
    'Licensed under the Apache License, Version 2.0 (the "License");',
    "you may not use this file except in compliance with the License.",
    "You may obtain a copy of the License at",
    "",
    "   http://www.apache.org/licenses/LICENSE-2.0",
    "",
    "Unless required by applicable law or agreed to in writing,",
    'software distributed under the License is distributed on an "AS IS" BASIS,',
    "WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.",
    "See the License for the specific language governing permissions and",
    "limitations under the License.",
]
NO_COPYRIGHT_LINE = "neuralmagic: no copyright"
QUALITY_COMMAND = "quality"
STYLE_COMMAND = "style"


def parse_args():
    """
    Setup and parse command line arguments for using the script
    """
    parser = argparse.ArgumentParser(
        description=(
            "Add Neuralmagic copyright to the beginning of all "
            "files under the given glob patterns. "
            "Currently assumes Python files using '#' as the commenting prefix."
        )
    )
    subparsers = parser.add_subparsers(dest="command")
    quality_parser = subparsers.add_parser(
        QUALITY_COMMAND,
        description=(
            "Run check across the files in the given patterns and "
            "fail if any do not have a copyright in them"
        ),
    )
    style_parser = subparsers.add_parser(
        STYLE_COMMAND,
        description=(
            "Add the copyright to any files in the given patterns if it is not present"
        ),
    )

    for sub in [quality_parser, style_parser]:
        sub.add_argument(
            "patterns",
            type=str,
            default=[],
            nargs="+",
            help="the patterns to search through",
        )

    return parser.parse_args()


def quality(patterns: List[str]):
    """
    Run a quality check across all files in the given glob patterns.
    This checks to make sure all matching files have the NM copyright present.
    If any do not, it will list them out and exit with an error.

    :param patterns: The glob file patterns to run quality check on
    """
    check_files = _get_files(patterns)
    error_files = []

    for file in check_files:
        if not _dont_copyright(file) and not _contains_copyright(file):
            print(f"would add copyright to {file}")
            error_files.append(file)

    if error_files:
        sys.exit(
            f"{len(error_files)} would be copyrighted, "
            f"{len(check_files) - len(error_files)} would be left unchanged."
        )
    else:
        print(f"{len(check_files)} files have copyrights")


def style(patterns: List[str]):
    """
    Run a style application across all files in the given glob patterns.
    This checks to make sure all matching files have the NM copyright present.
    If any do not, it will append the copyright to above the file after
    any already contained headers such as shebang lines.

    :param patterns: The glob file patterns to run quality check on
    """
    check_files = _get_files(patterns)
    copyrighted_files = []

    for file in check_files:
        if not _dont_copyright(file) and not _contains_copyright(file):
            _add_copyright(file)
            print(f"copyrighted {file}")
            copyrighted_files.append(file)

    if copyrighted_files:
        print(
            f"{len(copyrighted_files)} file(s) copyrighted, "
            f"{len(check_files) - len(copyrighted_files)} files unchanged"
        )
    else:
        print(f"{len(check_files)} files unchanged")


def _get_files(patterns: List[str]) -> List[str]:
    files = []

    for pattern in patterns:
        for file in glob.glob(pattern, recursive=True):
            files.append(os.path.abspath(os.path.expanduser(file)))

    files.sort()

    return files


def _dont_copyright(file_path: str) -> bool:
    with open(file_path, "r") as file:
        content = file.read()

    try:
        content.index(NO_COPYRIGHT_LINE)

        return True
    except ValueError:
        return False


def _contains_copyright(file_path: str) -> bool:
    with open(file_path, "r") as file:
        content = file.read()

    try:
        for line in COPYRIGHT_LINES:
            content.index(line)

        return True
    except ValueError:
        return False


def _add_copyright(file_path: str):
    file_type = _file_type(file_path)

    if file_type == "unknown":
        raise ValueError(
            f"unsupported file_type given to be copyrighted at {file_path}"
        )

    with open(file_path, "r+") as file:
        lines = file.readlines()
        header_info = _file_header_info(lines, file_type)
        inject_index = 0

        if header_info.end_index > -1:
            # if there is already a header, we want to inject the copyright after it
            # additionally we'll need a new line between the prev header and copyright
            inject_index = header_info.end_index + 1
            lines.insert(inject_index, "\n")
            inject_index += 1

        # add the copyright at the inject index
        file_copyright = _file_copyright(file_type)
        lines.insert(inject_index, file_copyright)

        if not header_info.new_line_after:
            # if there wasn't a new line after the header,
            # add in a new line after to create space between the code and copyright
            inject_index += 1
            lines.insert(inject_index, "\n")

        file.seek(0)
        file.writelines(lines)
        file.truncate()


def _file_copyright(file_type: str) -> str:
    comment_formatting = _code_comment_formatting(file_type)
    lines = []

    if comment_formatting.block_prefix:
        lines.append(comment_formatting.block_prefix)

    for line in COPYRIGHT_LINES:
        lines.append(
            f"{comment_formatting.line_prefix} {line}"
            if comment_formatting.line_prefix
            else line
        )

    if comment_formatting.block_suffix:
        lines.append(comment_formatting.block_suffix)

    # make sure there is a new line after last line of the copyright
    lines.append("")

    return "\n".join(lines)


_HeaderInfo = NamedTuple(
    "HeaderInfo",
    [
        ("start_index", int),
        ("end_index", int),
        ("new_line_before", bool),
        ("new_line_after", bool),
    ],
)


def _file_header_info(lines: List[str], file_type: str) -> _HeaderInfo:
    start_index = -1
    end_index = -1
    new_line_before = False
    new_line_after = False

    comment_formatting = _code_comment_formatting(file_type)
    prefix_found = False
    suffix_found = False

    for index, line in enumerate(lines):
        line = line.strip()

        if not line:
            # empty line, record the state of new lines before and after header
            if not prefix_found:
                new_line_before = True
            elif prefix_found and (suffix_found or not comment_formatting.block_suffix):
                new_line_after = True
        elif (
            comment_formatting.block_prefix
            and line.startswith(comment_formatting.block_prefix)
        ) or (
            not comment_formatting.block_prefix
            and line.startswith(comment_formatting.line_prefix)
        ):
            # start of header
            prefix_found = True
            start_index = index
            end_index = index
            suffix_found = comment_formatting.block_suffix and line.endswith(
                comment_formatting.block_suffix
            )
        elif comment_formatting.block_suffix and line.endswith(
            comment_formatting.block_suffix
        ):
            # end of header
            suffix_found = True
            end_index = index
        elif prefix_found and comment_formatting.block_suffix and not suffix_found:
            # in the middle of the header, searching for the end
            # reset new_line_after in case there was a break in the header
            new_line_after = True
        else:
            # first non header line, break out
            break

    return _HeaderInfo(start_index, end_index, new_line_before, new_line_after)


_CommentFormatting = NamedTuple(
    "CommentFormatting",
    [
        ("line_prefix", str),
        ("block_prefix", str),
        ("block_suffix", str),
    ],
)


def _code_comment_formatting(file_type: str) -> _CommentFormatting:
    if file_type == "python":
        return _CommentFormatting("#", "", "")
    elif file_type == "html" or file_type == "markdown":
        return _CommentFormatting("", "<!--", "-->")
    elif file_type == "css" or file_type == "javascript":
        return _CommentFormatting("", "/*", "*/")
    elif file_type == "restructuredtext":
        return _CommentFormatting("   ", "..", "")

    raise ValueError(f"unsupported file_type given for code prefix suffix: {file_type}")


def _file_type(file_path: str) -> str:
    if file_path.endswith(".py"):
        return "python"
    elif (
        file_path.endswith(".js")
        or file_path.endswith(".jsx")
        or file_path.endswith(".ts")
        or file_path.endswith(".tsx")
        or file_path.endswith(".jss")
    ):
        return "javascript"
    elif file_path.endswith(".html"):
        return "html"
    elif file_path.endswith(".css"):
        return "css"
    elif file_path.endswith(".md"):
        return "markdown"
    elif file_path.endswith(".rst"):
        return "restructuredtext"

    return "unknown"


def main():
    args = parse_args()

    if args.command == QUALITY_COMMAND:
        quality(args.patterns)
    elif args.command == STYLE_COMMAND:
        style(args.patterns)
    else:
        raise ValueError(f"unknown command given: {args.command}")


if __name__ == "__main__":
    main()


# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, Generator, Tuple

import numpy as np
import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationStrategy,
)
from compressed_tensors.quantization.lifecycle.forward import quantize
from compressed_tensors.utils import (
    get_permutations_24,
    is_quantization_param,
    merge_names,
    sparse_semi_structured_from_dense_cutlass,
    tensor_follows_mask_structure,
)
from torch import Tensor
from tqdm import tqdm


_LOGGER: logging.Logger = logging.getLogger(__name__)


@BaseCompressor.register(name=CompressionFormat.marlin_24.value)
class Marlin24Compressor(BaseCompressor):
    """
    Compresses a quantized model with 2:4 sparsity structure for inference with the
    Marlin24 kernel. Decompression is not implemented for this compressor.
    """

    @staticmethod
    def validate_quant_compatability(
        names_to_scheme: Dict[str, QuantizationScheme]
    ) -> bool:
        """
        Checks if every quantized module in the model is compatible with Marlin24
        compression. Quantization must be channel or group strategy with group_size
        of 128. Only symmetric quantization is supported

        :param names_to_scheme: dictionary of mapping module names to their
            quantization schemes
        :return: True if all modules are compatible with Marlin24 compression, raises
            a ValueError otherwise
        """
        for name, scheme in names_to_scheme.items():
            quant_args = scheme.weights
            if quant_args is None:
                raise ValueError(
                    "Marlin24 Compressor is only valid for weight quantization schemes"
                )

            strategy = quant_args.strategy
            group_size = quant_args.group_size
            symmetric = quant_args.symmetric
            if (
                strategy is not QuantizationStrategy.GROUP.value
                and strategy is not QuantizationStrategy.CHANNEL.value
            ):
                raise ValueError(
                    f"Marlin24 Compressor is only valid for group and channel "
                    f"quantization strategies, got {strategy} in {name}"
                )

            if group_size is not None and group_size != 128:
                raise ValueError(
                    f"Marlin24 Compressor is only valid for group size 128, "
                    f"got {group_size} in {name}"
                )

            if not symmetric:
                raise ValueError(
                    f"Marlin24 Compressor is only valid for symmetric quantzation, "
                    f"got symmetric={symmetric} in {name}"
                )

        return True

    @staticmethod
    def validate_sparsity_structure(name: str, weight: Tensor) -> bool:
        """
        Checks if a tensor fits the required 2:4 sparsity structure

        :param name: name of the tensor to check
        :param weight: tensor to check for sparsity structure
        :return: True if all rows match the 2:4 sparsity structure, raises
            ValueError otherwise
        """

        if not tensor_follows_mask_structure(weight):
            raise ValueError(
                "Marlin24 Compressor is only compatible with weights that have "
                f"a 2:4 sparsity structure. Found segments in {name} "
                "that do not match the expected structure."
            )

        return True

    @property
    def compression_param_names(self) -> Tuple[str]:
        """
        Returns a tuple of compression parameter names introduced by
        the compressor during compression
        """
        return ("weight_packed", "scale_packed", "meta")

    def compress(
        self,
        model_state: Dict[str, Tensor],
        names_to_scheme: Dict[str, QuantizationScheme],
        show_progress: bool = False,
        **kwargs,
    ) -> Dict[str, Tensor]:
        """
        Compresses a quantized state_dict with 2:4 sparsity structure for inference
        with the Marlin24 kernel

        :param model_state: state dict of uncompressed model
        :param names_to_scheme: quantization scheme for each quantized weight, needed
            for quantize function to calculate bit depth
        :param show_progress: whether to show tqdm progress
        :return: compressed state dict
        """
        self.validate_quant_compatability(names_to_scheme)

        compressed_dict = {}
        weight_suffix = ".weight"
        _LOGGER.debug(
            f"Compressing model with {len(model_state)} parameterized layers..."
        )

        for name, value in tqdm(
            model_state.items(), desc="Compressing model", disable=(not show_progress)
        ):
            if name.endswith(weight_suffix):
                prefix = name[: -(len(weight_suffix))]
                scale = model_state.get(merge_names(prefix, "weight_scale"), None)
                zp = model_state.get(merge_names(prefix, "weight_zero_point"), None)
                if scale is not None:  # weight is quantized, compress it

                    # Marlin24 kernel requires float16 inputs
                    scale = scale.to(torch.float16)
                    value = value.to(torch.float16)

                    # quantize weight, keeping it as a float16 for now
                    quant_args = names_to_scheme[prefix].weights
                    value = quantize(
                        x=value, scale=scale, zero_point=zp, args=quant_args
                    )

                    # compress based on sparsity structure
                    self.validate_sparsity_structure(prefix, value)
                    value, meta = compress_weight_24(value)
                    meta = meta.cpu()

                    # Marlin24 kernel expects input dim first
                    value = value.t().contiguous().cpu()
                    scale = scale.t().contiguous().cpu()
                    og_weight_shape = value.shape

                    # Marlin24 kernel expects unsigned values, shift zero-point
                    value += (1 << quant_args.num_bits) // 2

                    # pack quantized weight and scale
                    value = pack_weight_24(value, quant_args)
                    packed_scale = pack_scales_24(scale, quant_args, og_weight_shape)
                    meta = meta.resize_(meta.shape[1] // 2, meta.shape[0] * 2)

                    # save compressed values
                    compressed_dict[merge_names(prefix, "scale_packed")] = packed_scale
                    compressed_dict[merge_names(prefix, "weight_packed")] = value
                    compressed_dict[merge_names(prefix, "meta")] = meta
                    continue

            if not is_quantization_param(name):
                # export unquantized parameters without modifying
                compressed_dict[name] = value.to("cpu")

        return compressed_dict

    def decompress(
        self, path_to_model_or_tensors: str, device: str = "cpu", **kwargs
    ) -> Generator[Tuple[str, Tensor], None, None]:
        raise NotImplementedError(
            "Decompression is not implemented for the Marlin24 Compressor."
        )


def compress_weight_24(weight: Tensor):
    weight = weight.contiguous()
    w_comp, meta = sparse_semi_structured_from_dense_cutlass(weight)
    w_comp = w_comp.contiguous()
    return w_comp, meta


def marlin_permute_weights(q_w, size_k, size_n, perm, tile):
    assert q_w.shape == (size_k, size_n)
    assert size_k % tile == 0, f"size_k = {size_k}, tile = {tile}"
    assert size_n % tile == 0, f"size_k = {size_n}, tile = {tile}"

    # Permute weights to 16x64 marlin tiles
    q_w = q_w.reshape((size_k // tile, tile, size_n // tile, tile))
    q_w = q_w.permute((0, 2, 1, 3))
    q_w = q_w.reshape((size_k // tile, size_n * tile))

    q_w = q_w.reshape((-1, perm.numel()))[:, perm].reshape(q_w.shape)

    return q_w


def pack_weight_24(
    weight: Tensor,
    quantization_args: QuantizationArgs,
    tile: int = 16,
) -> torch.Tensor:
    size_k = weight.shape[0]
    size_n = weight.shape[1]
    num_bits = quantization_args.num_bits
    pack_factor = 32 // num_bits

    # Reshuffle to marlin_24 format
    perm, _, _ = get_permutations_24(num_bits)
    q_w = marlin_permute_weights(weight, size_k, size_n, perm, tile)

    q_w = q_w.cpu().numpy().astype(np.uint32)

    q_packed = np.zeros((q_w.shape[0], q_w.shape[1] // pack_factor), dtype=np.uint32)
    for i in range(pack_factor):
        q_packed |= q_w[:, i::pack_factor] << num_bits * i

    q_packed = torch.from_numpy(q_packed.astype(np.int32))

    return q_packed


def pack_scales_24(
    scales: torch.Tensor, quantization_args: QuantizationArgs, w_shape: torch.Size
) -> torch.Tensor:
    size_k = w_shape[0]
    size_n = w_shape[1]
    num_bits = quantization_args.num_bits

    _, scale_perm_2_4, scale_perm_single_2_4 = get_permutations_24(num_bits)

    if (
        quantization_args.strategy == QuantizationStrategy.GROUP
        and quantization_args.group_size < size_k
    ):
        scales = scales.reshape((-1, len(scale_perm_2_4)))[:, scale_perm_2_4]
    else:  # channelwise
        scales = scales.reshape((-1, len(scale_perm_single_2_4)))[
            :, scale_perm_single_2_4
        ]
    scales = scales.reshape((-1, size_n)).contiguous()

    return scales
